/* Programm für Schlupfwächter
 * Baierend auf AVRCDC von Osamu Tamura, AVRUSB von www.obdev.at
 * Hardware:
 * Der Schlupfwächter hat zwei differenzielle Eingänge,
 * die an Timer0 und Timer1 angeschlossen sind.
 * Timer2 dient als Zeitgeber für die Frequenzmessung (svw. Drehzahl)
 * Bei Feststellung von Schlupf schalten 2 Relais mit Zeitversatz ab.
 * Alle Ausgaben erfolgen derart, dass man HyperTerminal zur
 * Prozesüberwachung verwenden kann.
 * Eine Zweifarb-LED dient zur Kontrollanzeige:
 * grün = OK, rot = Ausfall
 *
 * tabsize = 8, encoding = utf-8 (ohne Windows-Header BOM)
 */

#include <string.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>

#include "usbdrv.h"

/************
 * Hardware *
 ************/
/*
Verwendung der Ports:
PB3/OC2	(15)	LED gemeinsame Kathode (Helligkeitsmodulation)

PC0	(23)	Relais 1 (low-aktiv)
PC1	(24)	Relais 2 (low-aktiv)
PC2	(25)	USB-Spannungskontrolle (Pulldown fehlt!)
PC4	(27)	LED grün Anode
PC5	(28)	LED rot Anode

PD2	(32)	USB D+
PD3	(1)	USB D-
PD4/T0	(2)	Eingang Puls2 (Komparator OPA347 vorgeschaltet)
PD5	(9)	Jumper low-aktiv für USB-Bootloader
PD6/AIN0 (10)	Eingang Puls1 +
PD7/AIN1 (11)	Eingang Puls1 -

Verwendung der Zähler:
Timer0 (8 bit) = Impulszähler, vorzugsweise für Geber mit wenig Strichen
Timer1 (16bit) = Impulszähler, vorzugsweise für Geber mit viel Strichen
Timer2 (8 bit) = Zeitgeber (Quarzfrequenzteiler), dazu PWM für LED
		 Taktung: 93,75 kHz = 10,6 µs (12 MHz / 128)
		 Überlauf-Interrupt: 366 Hz = 2,731 ms
		 Gesamtüberlauf (32 bit): 12 h
Alle 3 Zähler sind softwaremäßig auf 32 bit erweitert.

*/
#define CPUCLK 12000000	// Hz
#define T2PRE 128

typedef unsigned long ulong;
typedef unsigned short ushort;

uchar TimerHigh[3+2+3];	// High-Bytes für die 3 Timer

register uchar IrqSave asm("r2");	// für schnelle Interruptbehandlung
register uchar IrqSaveL asm("r3");

ISR(TIMER0_OVF_vect) __attribute__((naked));
ISR(TIMER0_OVF_vect) {
 asm volatile(
"	in	r2,__SREG__	\n"
"	lds	r3,TimerHigh+0	\n"
"	inc	r3		\n"
"	sts	TimerHigh+0,r3	\n"
"	brne	ee0		\n"
"	lds	r3,TimerHigh+1	\n"
"	inc	r3		\n"
"	sts	TimerHigh+1,r3	\n"
"	brne	ee0		\n"
"	lds	r3,TimerHigh+2	\n"
"	inc	r3		\n"
"	sts	TimerHigh+2,r3	\n"
"ee0:	out	__SREG__,r2	\n"
"	reti			\n");
}

ISR(TIMER1_OVF_vect) __attribute__((naked));
ISR(TIMER1_OVF_vect) {
 asm volatile(
"	in	r2,__SREG__	\n"
"	lds	r3,TimerHigh+3	\n"
"	inc	r3		\n"
"	sts	TimerHigh+3,r3	\n"
"	brne	ee1		\n"
"	lds	r3,TimerHigh+4	\n"
"	inc	r3		\n"
"	sts	TimerHigh+4,r3	\n"
"ee1:	out	__SREG__,r2	\n"
"	reti			\n");
}

ISR(TIMER2_OVF_vect) __attribute__((naked));
ISR(TIMER2_OVF_vect) {
 asm volatile(
"	in	r2,__SREG__	\n"
"	lds	r3,TimerHigh+5	\n"
"	inc	r3		\n"
"	sts	TimerHigh+5,r3	\n"
"	brne	ee2		\n"
"	lds	r3,TimerHigh+6	\n"
"	inc	r3		\n"
"	sts	TimerHigh+6,r3	\n"
"	brne	ee2		\n"
"	lds	r3,TimerHigh+7	\n"
"	inc	r3		\n"
"	sts	TimerHigh+7,r3	\n"
"ee2:	out	__SREG__,r2	\n"
"	reti			\n");
}

// Schanppschuss aller 3 Timer abholen
// Die Timer sind nicht übergelaufen, und die Timer-Interrups gesperrt.
// Der Rückgabewert enthält alle 3 Timer in 4 Bytes lückenlos verpackt.
// Kein cli(), um AVRUSB-Interruptlatenz zu garantieren.
static ulong GetTimerSnap(void) {
 ulong t;
 for (;;) {		// Hier beginnt der „Spinlock“
  TIMSK=0;		// alle Zählerinterrupts sperren
  asm volatile(
"	in	%A0,0x32	\n"	// TCNT0
"	in	%B0,0x2C	\n"	// TCNT1L
"	in	%C0,0x2D	\n"	// TCNT1H
"	in	%D0,0x24	\n"	// TCNT2
:"=r" (t)
:
  );
  if (!(TIFR&0x45)) break; // inzwischen irgendwelche Überläufe?
  TIMSK=0x45;		// ja, ISRs abarbeiten lassen, dann neuer Versuch
 }
// Jetzt können die Zähler ruhig weiterzählen und auch überlaufen
 return t;
}
 
// Schnappschüsse der drei Timer holen ohne überlaufeffekte
static void GetTimers(ulong values[3]) {
 ulong t=GetTimerSnap();	// Timer-Stände überlauffrei abholen
 asm volatile(			// mit High-Bytes kombinieren (schnell!)
"	st	%a2+,%A0	\n"	// Timer0
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	st	%a2+,%B0	\n"	// Timer1
"	st	%a2+,%C0	\n"
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	st	%a2+,%D0	\n"	// Timer2
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
"	ld	r0,%a1+		\n"
"	st	%a2+,r0		\n"
:					// kein Ausgaberegister
:"r" (t)				// 3 Eingaben
,"e" (TimerHigh)
,"e" (values)
:"memory"				// Speicher wird verändert
 );
// Jetzt können die Überläufe zum Erhöhen der High-Teile verwendet werden
 TIMSK=0x45;			// Timer-Interupts freigeben
}

/*******
 * LED *
 *******/
static uchar Led_T;	// flashing time in ms, 0 = LED is steady
static uchar Led_F;	// flashing frequency in ms (half period)
static uchar Led_C;	// flash-counter (via SOF pulses)

void Led_Start(uchar t) {
 if (!Led_T) {
  Led_C=t;
  DDRB&=!0x08;		// LED ausschalten
 }
 Led_T=Led_F=t;
}

static void Led_On1ms(void) {
 uchar led_t=Led_T, led_c;	// lokal vorhalten
 if (!led_t) return;
 led_c=Led_C;
 if (!--led_c) {
  DDRB^=0x08;			// LED ein- oder ausschalten
  led_c=Led_F;
 }
 if (!--led_t) DDRB|=0x08;	// LED einschalten
 Led_T=led_t;
 Led_C=led_c;
}

typedef struct {		// Zu speichernde Daten (EEPROM)
 unsigned Strich1, Strich2;	// Strichzahl der Geber pro Umdrehung
 unsigned Zahn1, Zahn2;		// Zähnezahl oder Umfang (Durchmesser)
 float MaxSchlupf;
 unsigned Rel2Delay;		// Abfallverzögerung für 2. Relais [ms]
 uchar Synchron;		// !=0 bei Synchrongetriebe, ==0 asynchron
}persistent_t;
/* Definition von Schlupf bzw. MaxSchlupf:
   1. Asynchroner Fall (Drehzahl2*Zahn2 ≥ Drehzahl1*Zahn1):
   Schlupf = 1 - (Drehzahl1*Zahn1) : (Drehzahl2*Zahn2), hier 0 ≤ Schlupf ≤ 1
   2. Synchroner Fall: Schlupf = Anzahl „übersprungene“ Zähne, keine Grenze
   Schlupf = Position1*Zahn1 - Position2*Zahn2
*/

/***********************************************
 * Einfacher Ringpuffer (ohne Interruptsperre) *
 ***********************************************/
typedef struct {
 uchar r,w,l,f;	// Lesezeiger, Schreibzeiger, Puffergröße (0==256), Flags
 uchar buf[0];	// Puffer (1..256 Bytes)
}ringbuf_t;
//Bei r==w kann die FIFO entweder leer oder voll sein.
#define EF 1	// Empty Flag
#define FF 2	// Full flag

// Datagramm in <buf> stecken, liefert !=0 wenn's _nicht_ passt
static uchar ringbuf_Put(ringbuf_t *buf, const uchar*data, uchar len) {
 register uchar r,w,l,f;
 if (!len) return len;	// nichts zu tun
 f=buf->f;
 if (f&FF) return len;	// Ringpuffer voll!
 r=buf->r;		// Daten lokal "cachen"
 w=buf->w;
 l=buf->l;
 if (r<=w) r+=l;	// egal, wenn's hier überläuft!
 r-=w;			// r = Platz, 0 = 256
 if (r && len>r) return len; // passt nicht
 do{
  buf->buf[w++]=*data++;
  if (w==l) w=0;
 }while (--len);
 f&=~EF;
 if (w==buf->r) f|=FF;
 buf->w=w;		// rückschreiben
 buf->f=f;
 return len;		// Datagramm passt (Null liefern)
}

// Daten aus Ringpuffer lesen, liefert Anzahl gelesener Bytes
static uchar ringbuf_Get(ringbuf_t *buf, uchar*data, uchar len) {
 uchar ret=0, r,w,l,f;
 f=buf->f;
 if (f&EF) return ret;	// Puffer leer
 r=buf->r;		// cachen
 w=buf->w;
 l=buf->l;
 f&=~FF;
 do{
  *data++=buf->buf[r++];
  if (r==l) r=0;
  ret++;
  if (r==w) {f|=EF; break;}
 }while (--len);
 buf->r=r;		// rückschreiben
 buf->f=f;
 return ret;
}

/*******************************************
 * Kooperatives Multithreading (2 Threads) *
 *******************************************/

uchar *OtherSP;		// Der jeweils andere Stackpointer

// Threadumschaltung
#define NUMPUSH 16	// 16 Bytes werden hier gePUSHt
void Yield(void) __attribute__((naked));
void Yield(void) {
 asm volatile(
"	push	r4		\n"
"	push	r5		\n"
"	push	r6		\n"
"	push	r7		\n"
"	push	r8		\n"
"	push	r9		\n"
"	push	r10		\n"
"	push	r11		\n"
"	push	r12		\n"
"	push	r13		\n"
"	push	r14		\n"
"	push	r15		\n"
"	push	r16		\n"
"	push	r17		\n"
"	push	r28		\n"
"	push	r29		\n"
"	in	r16,__SP_L__	\n"
"	in	r17,__SP_H__	\n"
"	ld	r28,%a0		\n"
"	st	%a0+,r16	\n"
"	ld	r29,%a0		\n"
"	st	%a0+,r17	\n"
"	cli			\n"
"	out	__SP_L__,r28	\n"
"	out	__SP_H__,r29	\n"
"	sei			\n"
"	pop	r29		\n"	// vom anderen Stack!
"	pop	r28		\n"
"	pop	r17		\n"
"	pop	r16		\n"
"	pop	r15		\n"
"	pop	r14		\n"
"	pop	r13		\n"
"	pop	r12		\n"
"	pop	r11		\n"
"	pop	r10		\n"
"	pop	r9		\n"
"	pop	r8		\n"
"	pop	r7		\n"
"	pop	r6		\n"
"	pop	r5		\n"
"	pop	r4		\n"
"	ret			\n"
::"e" (&OtherSP)
 );
}

/*********************************
 * Puffer für seriellen Transfer *
 *********************************/
 
static struct{
 ringbuf_t ring;
 uchar buf[32];
}recvbuf;

static struct{
 ringbuf_t ring;
 uchar buf[256];
}sendbuf;

// Routine blockiert (schaltet Thread) bis Zeichen ausgegeben
extern int fputc(int c, FILE *f) {
 static uchar buf;	// Der Compiler ist zu doof seine Register zu adressieren!
 buf=(uchar)c;
 while (ringbuf_Put(&sendbuf.ring,&buf,1)) Yield();
 return (uchar)c;
}

// Routine blockiert nie, liefert EOF wenn kein Zeichen vorhanden
extern int fgetc(FILE *f) {
 static uchar c;
 if (ringbuf_Get(&recvbuf.ring,&c,1)) return c;
 return -1;
}

static uchar inchar;

// Tastenkode abholen, liefert 0 wenn keine Taste angeschlagen.
// HyperTerminal-Sondertasten werden umgewandelt entsprechend
// http://www-user.tu-chemnitz.de/~heha/hs_freeware/terminal/terminal.htm#3.7
static uchar peekchar(void) {
 static uchar EscapeMode;
 if (ringbuf_Get(&recvbuf.ring,&inchar,1)) {
  if (EscapeMode) {
   switch (inchar&0x7F) {
    case 0x1B: inchar='C'-'@'; break;		// 2x Escape = ^C
    case '?':
    case '[':
    case 'O': inchar=0; goto keepescape;	// komplett ignorieren
    case 'A': inchar='P'-'@'; break;		// ↑
    case 'B': inchar='N'-'@'; break;		// ↓
    case 'C': inchar='F'-'@'; break;		// →
    case 'D': inchar='B'-'@'; break;		// ←
    case 'H': inchar='A'-'@'; break;		// Pos1
    case 'K': inchar='E'-'@'; break;		// Ende
    case 'P': inchar=0x1C; break;		// F1
    case 'Q': inchar=0x1D; break;		// F2
    case 'R': inchar=0x1E; break;		// F3
    case 'S': inchar=0x1F; break;		// F4
    default: inchar=0;				// unbekannt: verwerfen!
   }
   EscapeMode=0;
keepescape:;
  }else if (inchar==0x1B) {
   EscapeMode=inchar;
   inchar=0;
  }
 }else inchar=0;
 return inchar;
}

/*******************
 * Frequenzmessung *
 *******************/
#define MEAS_MIN_TIME 100	// Millisekunden

static persistent_t ramState;
 
static ulong DiffTimers[3];
static ulong LastTimers[3];
static ulong CurrTimers[3];

static void PeriodicAction(void) {
 GetTimers(CurrTimers);
 DiffTimers[2]=CurrTimers[2]-LastTimers[2];
 if (DiffTimers[2]<MEAS_MIN_TIME*CPUCLK/T2PRE/1000) return;	// keine Aktion
 DiffTimers[0]=CurrTimers[0]-LastTimers[0];
 DiffTimers[1]=CurrTimers[1]-LastTimers[1];
// Nur bei asynchronem Betrieb!
 memcpy(LastTimers,CurrTimers,sizeof(LastTimers));
}

void __attribute__((noreturn)) Thread2Entry(void) {
start:
 PORTC^=0x20;
 printf_P(PSTR("\e[37;40m" "\e[2J" "\e[H" "Hallo Welt\n"));
 for(;;) {
  if (peekchar()) switch (inchar) {
   case 'L'-'@': goto start;
   case 't': printf_P(PSTR("Test\n")); break;
   case 'T': ringbuf_Put(&sendbuf.ring,(const uchar*)"Teee\r\n",6); break;
   default: putchar(inchar); break;
  }
  Yield();
 }
}

/*********************
 * USB and CDC stuff *
 *********************/

enum {
 SEND_ENCAPSULATED_COMMAND = 0,
 GET_ENCAPSULATED_RESPONSE,
 SET_COMM_FEATURE,
 GET_COMM_FEATURE,
 CLEAR_COMM_FEATURE,
 SET_LINE_CODING = 0x20,
 GET_LINE_CODING,
 SET_CONTROL_LINE_STATE,
 SEND_BREAK};

PROGMEM char usbDescriptorConfiguration[] = {   /* USB configuration descriptor */
 9,	/* sizeof(usbDescrConfig): length of descriptor in bytes */
 USBDESCR_CONFIG,	/* descriptor type */
 67,0,	/* total length of data returned (including inlined descriptors) */
 2,	/* number of interfaces in this configuration */
 1,	/* index of this configuration */
 0,	/* configuration name string index */
 USBATTR_SELFPOWER,	/* attributes */
 0,	/* max USB current in 2mA units */

/* interface descriptor */
 9,	/* sizeof(usbDescrInterface): length of descriptor in bytes */
 USBDESCR_INTERFACE, /* descriptor type */
 0,	/* index of this interface */
 0,	/* alternate setting for this interface */
 1,	/* endpoints excl 0: number of endpoint descriptors to follow */
 2,	/* control interface class: CDC */
 2,	/* control interface subclass: Abstract (Modem) */
 1,	/* control interface protocol: AT commands */
 0,	/* string index for interface */

/* CDC Class-Specific descriptor */
 5,	/* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
 0x24,	/* descriptor type */
 0,	/* header functional descriptor */
 0x10,0x01,

 4,	/* sizeof(usbDescrCDC_AcmFn): length of descriptor in bytes */
 0x24,	/* descriptor type */
 2,	/* abstract control management functional descriptor */
 0x02,	/* SET_LINE_CODING,    GET_LINE_CODING, SET_CONTROL_LINE_STATE    */

 5,	/* sizeof(usbDescrCDC_UnionFn): length of descriptor in bytes */
 0x24,	/* descriptor type */
 6,	/* union functional descriptor */
 0,	/* CDC_COMM_INTF_ID */
 1,	/* CDC_DATA_INTF_ID */

 5,	/* sizeof(usbDescrCDC_CallMgtFn): length of descriptor in bytes */
 0x24,	/* descriptor type */
 1,	/* call management functional descriptor */
 3,	/* allow management on data interface, handles call management by itself */
 1,	/* CDC_DATA_INTF_ID */

/* Endpoint Descriptor */
 7,	/* sizeof(usbDescrEndpoint) */
 USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
 0x83,	/* IN endpoint number 3 */
 0x03,	/* attrib: Interrupt endpoint */
 8,0,	/* maximum packet size */
 2,	/* polling interval in ms */

/* Second interface Descriptor */
 9,	/* sizeof(usbDescrInterface): length of descriptor in bytes */
 USBDESCR_INTERFACE,           /* descriptor type */
 1,	/* index of this interface */
 0,	/* alternate setting for this interface */
 2,	/* endpoints excl 0: number of endpoint descriptors to follow */
 0x0A,	/* Data Interface Class */
 0,	/* Data Interface Subclass */
 0,	/* Data Interface Class Protocol Codes */
 0,	/* string index for interface */

/* Endpoint Descriptor */
 7,           /* sizeof(usbDescrEndpoint) */
 USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
 0x01,	/* OUT endpoint number 1 */
 0x02,	/* attrib: Bulk endpoint */
 8,0,	/* maximum packet size 8->6 */
 0,

/* Endpoint Descriptor */
 7,           /* sizeof(usbDescrEndpoint) */
 USBDESCR_ENDPOINT,  /* descriptor type = endpoint */
 0x81,	/* IN endpoint number 1 */
 0x02,	/* attrib: Bulk endpoint */
 8,0,	/* maximum packet size */
 0,
};



static uchar	requestType;
static uchar	modeBuffer[7];
static uchar	intr3Status;    /* used to control interrupt endpoint transmissions */
static uchar	lastLen;	// if 8, send an empty BULK packet

/***********************
 * SETUP data received *
 ***********************/
uchar usbFunctionSetup(uchar data[8]) {
 usbRequest_t *rq = (void*)data;

 if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {    /* class request type */
  if (rq->bRequest==GET_LINE_CODING || rq->bRequest==SET_LINE_CODING) {
   requestType = rq->bRequest;
   return 0xff;
/*    GET_LINE_CODING -> usbFunctionRead()    */
/*    SET_LINE_CODING -> usbFunctionWrite()    */
  }
  if (rq->bRequest == SET_CONTROL_LINE_STATE) {
/* Report serial state (carrier detect). On several Unix platforms,
 * tty devices can only be opened when carrier detect is set.
 */
   if (!intr3Status) intr3Status=2;
  }

/*  Prepare bulk-in endpoint to respond to early termination   */
  if ((rq->bmRequestType & USBRQ_DIR_MASK) == USBRQ_DIR_HOST_TO_DEVICE)
   lastLen=8;	// send an empty packet
 }
 return 0;
}

/***********************
 * IN transfer for EP0 *
 ***********************/
uchar usbFunctionRead(uchar *data, uchar len) {
 if (requestType == GET_LINE_CODING) {
  memcpy(data,modeBuffer,7);
  return 7;
 }
 return 0;	/* error -> terminate transfer */
}


/************************
 * OUT transfer for EP0 *
 ************************/
uchar usbFunctionWrite(uchar *data, uchar len) {
 if (requestType == SET_LINE_CODING) {
  memcpy(modeBuffer,data,7);
  return 1;
 }
 return 1;	/* error -> accept everything until end */
}

/***********************************
 * OUT-transfer für EP1 (Tastatur) *
 ***********************************/
 
void usbFunctionWriteOut(uchar *data, uchar len) {
 if (!len) return;
 Led_Start(150);		// fast flashing (3 Hz)
 ringbuf_Put(&recvbuf.ring,data,len);
}

/*************************************
 * Initialisierung und Hauptschleife *
 *************************************/
EEMEM persistent_t eeState;

uchar Stack2[256];

static void hardwareInit(void) {
 static FILE mystdout=FDEV_SETUP_STREAM((int(*)(char,FILE*))fputc,fgetc,_FDEV_SETUP_RW);
 PORTB=0xFF;		// Pullups EIN (wegen offener Eingänge)
 PORTC=0xDC;		// grüne LED und Relais EIN
 DDRC =0x33;
 PORTD=0x23;		// Pullup für Bootload-Jumper und freie Eingänge
 eeprom_read_block(&ramState,&eeState,sizeof(persistent_t));
 recvbuf.ring.l=sizeof(recvbuf.buf);
 recvbuf.ring.f=EF;
 sendbuf.ring.l=(uchar)sizeof(sendbuf.buf);
 sendbuf.ring.f=EF;
 TCCR0=0x07;		// Timer0: Taktquelle T0 (Flanke egal)
 TCCR1B=0x87;		// Timer1: Taktquelle T1 (Flanke egal), Spikefilter
 TCCR2=0x7B;		// Timer2: Taktquelle 93,75 kHz, schnelle PWM
 OCR2=0x80;		// LED: halbe Helligkeit
 TIMSK=0x45;		// Überlaufinterrupts für alle 3 Timer aktivieren
 Stack2[255]=((unsigned)Thread2Entry)&255;	// Lowbyte "oben"
 Stack2[254]=((unsigned)Thread2Entry)>>8;	// Highbyte "unten"
 OtherSP=Stack2+253-NUMPUSH;
// Yield() holt 16 Bytes vom Stack und macht dann RET
 stdout=&mystdout;
}

// Updates an ATmega-internal EEPROM location when different
static void ActualizeEepromByte(uchar *ee_addr, uchar b) {
 if (eeprom_read_byte(ee_addr)==b) return;
 EEDR=b;
 cli();
 EECR|=1<<EEMWE;
 EECR|=1<<EEWE;		// don't wait here!
 sei();
}

static void ActualizeEepromB(const uchar *pRam, uchar *pEe, size_t n) {
 if (!n) return;
 do{
  if (!eeprom_is_ready()) return;
  ActualizeEepromByte(pEe++,*pRam++);
 }while(--n);
}
#define ActualizeEeprom(r,e,l) ActualizeEepromB((const uchar*)r,(uchar*)e,l)

int __attribute__((noreturn)) main(void) {
 uchar SofCmp=0;
 MCUCR=0x80;		// SLEEP_MODE_STANDBY
 usbInit();
 sei();
 hardwareInit();

 for(;;){	// main event loop (infinite, possibly broken by watchdog)
  uchar t=usbSofCount;
  PeriodicAction();
  Yield();
  usbPoll();
  if (t-SofCmp) {
   SofCmp=t;
   Led_On1ms();
  }

  if (usbInterruptIsReady()) {
   uchar len;
   static uchar tbuf[8];
   len=ringbuf_Get(&sendbuf.ring,tbuf,8);
   if (len || lastLen&8 ) {
    usbSetInterrupt(tbuf,len);
    lastLen=len;
   }
  }

  if (usbInterruptIsReady3() && intr3Status) {
   static uchar serialStateNotification[10] = {
     0xa1, 0x20, 0, 0, 0, 0, 2, 0, 3, 0};

   if (intr3Status==2) usbSetInterrupt3(serialStateNotification, 8);
   else                usbSetInterrupt3(serialStateNotification+8, 2);
   intr3Status--;
  }
  ActualizeEeprom(&ramState,&eeState,sizeof(persistent_t));
 }
}
